"""
This library provides a set of utilities for detecting identity reverts in
revisioned content.

:Basic example:

    >>> import mwreverts
    >>>
    >>> checksum_revisions = [
    ...     ("aaa", {'rev_id': 1}),
    ...     ("bbb", {'rev_id': 2}),
    ...     ("aaa", {'rev_id': 3}),
    ...     ("ccc", {'rev_id': 4})
    ... ]
    >>>
    >>> list(mwreverts.detect(checksum_revisions))
    [Revert(reverting={'rev_id': 3},
            reverteds=[{'rev_id': 2}],
            reverted_to={'rev_id': 1})]

"""
from .detector import Detector, Revert
from .functions import detect
from .dummy_checksum import DummyChecksum

__version__ = "0.0.2"

__all__ = [Detector, Revert, detect, DummyChecksum]



RADIUS = 15
"""
Priedhorsky, R., Chen, J., Lam, S. T. K., Panciera, K., Terveen, L., &
Riedl, J. (2007, November). Creating, destroying, and restoring value in
Wikipedia. In Proceedings of the 2007 international ACM conference on
Supporting group work (pp. 259-268). ACM.
"""



from . import defaults
from .historical_dict import HistoricalDict
from .revert import Revert


class Detector(HistoricalDict):
    """
    Detects revert events in a stream of revisions (to the same page) based on
    matching checksums.  To detect reverts, construct an instance of this class
    and call :func:`~mwreverts.Detector.process` in chronological order.

    See https://meta.wikimedia.org/wiki/R:Identity_revert

    :Parameters:
        radius : int
            a positive integer indicating the maximum revision distance that a
            revert can span.

    :Example:
        >>> import mwreverts
        >>> detector = mwreverts.Detector()
        >>>
        >>> detector.process("aaa", {'rev_id': 1})
        >>> detector.process("bbb", {'rev_id': 2})
        >>> detector.process("aaa", {'rev_id': 3})
        Revert(reverting={'rev_id': 3},
               reverteds=[{'rev_id': 2}],
               reverted_to={'rev_id': 1})
        >>> detector.process("ccc", {'rev_id': 4})

    """

    def initialize(self, radius=defaults.RADIUS):
        if radius < 1:
            raise TypeError("invalid radius. Expected a positive integer.")

        super().initialize(maxsize=radius + 1)

    def process(self, checksum, revision=None):
        """
        Process a new revision and detect a revert if it occurred.  Note that
        you can pass whatever you like as `revision` and it will be returned in
        the case that a revert occurs.

        :Parameters:
            checksum : str
                Any identity-machable string-based hash of revision content
            revision : `mixed`
                Revision metadata.  Note that any data will just be returned
                in the case of a revert.

        :Returns:
            a :class:`~mwreverts.Revert` if one occured or `None`
        """
        revert = None

        if checksum in self:  # potential revert

            reverteds = list(self.up_to(checksum))

            if len(reverteds) > 0:  # If no reverted revisions, this is a noop
                revert = Revert(revision, reverteds, self[checksum])

        self.insert(checksum, revision)
        return revert



class DummyChecksum():
    """
    Used in when checking for reverts when the checksum of the revision of
    interest is unknown.  DummyChecksums won't match eachother or anything
    else, but they will match themselves and they are hashable.

    >>> dummy1 = DummyChecksum()
    >>> dummy1
    <#140687347334280>
    >>> dummy1 == dummy1
    True
    >>>
    >>> dummy2 = DummyChecksum()
    >>> dummy2
    <#140687347334504>
    >>> dummy1 == dummy2
    False
    >>>
    >>> {"foo", "bar", dummy1, dummy1, dummy2}
    {<#140687347334280>, 'foo', <#140687347334504>, 'bar'}
    """

    def __str__(self):
        repr(self)

    def __repr__(self):
        return "<#" + str(id(self)) + ">"



import os

from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def requirements(fname):
    for line in open(os.path.join(os.path.dirname(__file__), fname)):
        yield line.strip()


setup(
    name="mwreverts",
    version="0.0.2",  # Change in mwreverts/__init__.py
    author="Aaron Halfaker",
    author_email="aaron.halfaker@gmail.com",
    url="http://github.com/mediawiki-utilities/python-mwreverts",
    packages=["mwreverts"],
    license=read("LICENSE"),
    description="A set of utilities for detecting reverts in MediaWiki " +
                "revisions.",
    long_description=read("README.md"),
    install_requires=requirements("requirements.txt")
)



from collections import namedtuple

Revert = namedtuple("Revert", ['reverting', 'reverteds', 'reverted_to'])
"""
Represents a revert event.  This class behaves like
:class:`collections.namedtuple`.  Note that the datatypes of `reverting`,
`reverteds` and `reverted_to` is not specified since those types will depend
on the revision data provided during revert detection.

:Attributes:
    **reverting**
        The reverting revision data : `mixed`
    **reverteds**
        The reverted revision data (ordered chronologically) : list( `mixed` )
    **reverted_to**
        The reverted-to revision data : `mixed`
"""



from collections import deque

import jsonable


class HistoricalDict(jsonable.Type, dict):
    '''
    A datastructure for efficiently storing and retrieving a
    limited number of records based on keys.
    '''
    __slots__ = ('maxsize', 'history')

    def initialize(self, maxsize, history=None):
        '''size specifies the maximum amount of history to keep'''
        super().__init__()

        self.maxsize = int(maxsize)
        self.history = deque(maxlen=self.maxsize)  # Preserves order history

        # If `items` are specified, then initialize with them
        if history is not None:
            for key, value in history:
                self.insert(key, value)

    def __setitem__(self, key, value):
        self.insert(key, value)

    def insert(self, key, value):
        '''Adds a new key-value pair. Returns any discarded values.'''

        # Add to history and catch expectorate
        if len(self.history) == self.maxsize:
            expectorate = self.history[0]
        else:
            expectorate = None

        self.history.append((key, value))

        # Add to the appropriate list of values
        if key in self:
            super().__getitem__(key).append(value)
        else:
            super().__setitem__(key, [value])

        # Clean up old values
        if expectorate is not None:
            old_key, old_value = expectorate
            super().__getitem__(old_key).pop(0)
            if len(super().__getitem__(old_key)) == 0:
                super().__delitem__(old_key)

            return (old_key, old_value)

    def __getitem__(self, key):
        if key in self:
            return super().__getitem__(key)[-1]
        else:
            raise KeyError(key)

    def up_to(self, key):
        '''Gets the recently inserted values up to a key'''
        for okey, ovalue in reversed(self.history):
            if okey == key:
                break
            else:
                yield ovalue

    def last(self):
        return self.history[-1]

    def __eq__(self, other):
        if not hasattr(other, "history"):
            return False
        else:
            return self.history == other.history
